library ac_dart.server.app;

import 'acdart_router.dart';
import 'package:atlassian_connect_config/atlassian_connect_config.dart';

import 'package:shelf/shelf.dart';
import 'package:shelf/shelf_io.dart' as io;
import 'dart:async';
import 'package:shelf_route/shelf_route.dart';
import 'package:shelf_exception_response/exception_response.dart';
import 'package:logging/logging.dart';
import 'package:atlassian_connect_host/atlassian_connect_host.dart';
import 'dart:io';
import 'acdart_shelf_middleware.dart' as m show
  moustacheFile, productHostAuthenticator, sessionTokenAuthenticator,
  getSessionContext, staticFile, jwtTimingLogger;

export 'package:shelf_route/shelf_route.dart';
export 'package:shelf_bind/shelf_bind.dart';
export 'package:constrain/constrain.dart';

import 'package:shelf_static/shelf_static.dart' as static
  show createStaticHandler;

AcDartApp acdart(AcDartConfigFactory configFactory) =>
    new AcDartApp(configFactory);


class AcDartApp {
  final AcDartRouter router;
  final AcHostService hostService;
  final AcDartConfig config;
  final Middleware atlassianHostAuthenticator;
  final Middleware addonSessionAuthenticator;
  final _MoustacheFile moustacheFile = m.moustacheFile;
  final _StaticFile staticFile = m.staticFile;
  final Middleware jwtTimingLogger = m.jwtTimingLogger;
  bool _hasStarted = false;

  
  AcDartApp._internal(this.config, this.router, this.hostService,
      this.atlassianHostAuthenticator, this.addonSessionAuthenticator);
  
  factory AcDartApp(AcDartConfigFactory configFactory) {
    
    final config = configFactory.config;

    final addonBaseUrl = config.addonBaseUrl;

    final hostService = new AcHostService(config.acHostRepo, config.addonHost);

    final AcDartRouter rootRouter =
        new AcDartRouter(hostService, config.addonHost);
    
    return new AcDartApp._internal(config, rootRouter, hostService,
        m.productHostAuthenticator(hostService),
        m.sessionTokenAuthenticator(hostService));
  }
  
  // TODO: how to configure logging? Could actually move to AcDartConfig ??
//  void configureLogger() {
//  }
  
  void start() {
    if (_hasStarted) {
      return;
    }
    _hasStarted = true;
    
    _initLogging();
    
    printRoutes(router);

    var rootHandler = const Pipeline()
        .addMiddleware(logRequests(logger: _requestLogger)) // TODO: temp until can support in router
        .addMiddleware(exceptionResponse())
        .addHandler(router.handler);

    void _doAutoInstall() {
      _autoRegisterHostUrls.forEach((baseUrl) =>
          hostService.installAddonToHost(baseUrl));
    }
    
    io.serve(rootHandler, InternetAddress.ANY_IP_V6, config.serverPort)
      .then((server) {
        print('Serving at http://${server.address.host}:${server.port}');

        if (config.installAddonOnStartup) {
          return new Future(() {
            _log.info('installing addon...');
            _doAutoInstall();
          }).catchError((e, stackTrace) {
            _log.warning('caught error', e, stackTrace);
        });
        }

      })
      .catchError((e, stackTrace) {
        _log.warning('caught error', e, stackTrace);
    });

  }

  Handler createStaticHandler({ String fileSystemPath: '../web',
      bool serveFilesOutsidePath: false, String defaultDocument: 'index.html'}) {
    return static.createStaticHandler(fileSystemPath,
        serveFilesOutsidePath: serveFilesOutsidePath,
        defaultDocument: defaultDocument);
  }

}



void _initLogging() {
  Logger.root.level = Level.FINEST; // TODO: read from env
  Logger.root.onRecord.listen((LogRecord rec) {
    print('${rec.level.name}: ${rec.time}: ${rec.message}');
  });
}


// TODO: temp borrowed from router
void _requestLogger(String msg, bool isError) {
  if (isError) {
    _log.warning(msg);
  }
  else {
    _log.finest(msg);
  }
}

// TODO: move into config so can auto register for non prod herokus (assuming that works)
final List<Uri> _autoRegisterHostUrls = [
  Uri.parse("http://localhost:1990/confluence"),
  Uri.parse("http://localhost:2990/jira")
];

typedef Handler _MoustacheFile(String path, 
                         { Map<String, dynamic> templateVariables,
                           bool includeHeaders, 
                           bool includeExtraParams});

typedef Handler _StaticFile(String path);

final Logger _log = new Logger('AcDartServer');

